from typing import TYPE_CHECKING, get_args

from eth_rpc.types import Name
from eth_rpc.utils import is_annotation
from pydantic import ConfigDict

from .contract import Contract
from .func_signature import FuncSignature
from .function import ContractFunc


class _ProtocolBase(Contract):
    model_config = ConfigDict(extra="allow")

    def __init__(self, **kwargs):
        """
        This is how we make the ProtocolBase inject the ContractFunc via the type signature.
        After a lot of research, this goes slightly outside of the bounds for type hinting,
        but I think the improved expressiveness makes it fully worthwhile.
        """
        super().__init__(**kwargs)

        for alias, func in self._func_sigs.items():
            name = alias
            if is_annotation(func):
                annotation_args = get_args(func)
                args = annotation_args[0]
                for annotation in annotation_args:
                    if isinstance(annotation, Name):
                        name = annotation.value
            else:
                args = func
            T, U = get_args(args)

            setattr(
                self,
                alias,
                ContractFunc[T, U](  # type: ignore
                    func=FuncSignature[T, U](name=name, alias=alias),  # type: ignore
                    contract=self,
                ),
            )


if TYPE_CHECKING:
    # This is because type checkers don't like when you use __class_getitem__ outside of a generic
    # By adding a metaclass with a getitem, it plays much nicer with mypy.
    from pydantic._internal._model_construction import ModelMetaclass

    class IndexedKlass(ModelMetaclass):
        def __getitem__(self, item):
            return self

    class ProtocolBase(_ProtocolBase, metaclass=IndexedKlass):
        pass

else:
    ProtocolBase = _ProtocolBase



from collections.abc import Awaitable
from dataclasses import dataclass
from typing import Any, NewType, Optional, TypedDict, TypeVar, Union

from eth_typing import HexAddress, HexStr, TypeStr
from pydantic import BaseModel, Field

from . import primitives
from .args import (  # noqa: F401
    AlchemyBlockReceipt,
    AlchemyParams,
    BlockNumberArg,
    CallWithBlockArgs,
    EthCallArgs,
    EthCallParams,
    FeeHistoryArgs,
    GetAccountArgs,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
    GetCodeArgs,
    GetStorageArgs,
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    LogsArgs,
    LogsParams,
    RawTransaction,
    TraceArgs,
    TransactionRequest,
)
from .basic import (
    ALL_PRIMITIVES,
    BLOCK_STRINGS,
    BlockReference,
    Bytes32Hex,
    HexInt,
    HexInteger,
)
from .event import Indexed, Name
from .network import BlockExplorer, Network, Rpcs, RpcUrl
from .primitives import BASIC_TYPES, BYTES_TYPES
from .response import NoArgs, RPCResponseModel
from .struct import Struct
from .transaction import SignedTransaction
from .typed_data import Domain, EIP712Model, hash_eip712_bytes

T = TypeVar("T")
MaybeAwaitable = Union[T, Awaitable[T]]

METHOD = Field(init=False)
VersionNumber = NewType("VersionNumber", str)
EventType = NewType("EventType", BaseModel)


class ContractMethod:
    init: bool = False


@dataclass
class EvmTuple:
    values: list[Any]


class ABIParamType(BaseModel):
    indexed: bool = False
    name: str = ""
    type: TypeStr
    components: Optional[list["ABIParamType"]] = None

    def get_type(self):
        if self.components:
            return ",".join(component.get_type() for component in self.components)
        return self.type

    def model_dump(self, exclude_none=True, **kwargs):
        return super().model_dump(exclude_none=exclude_none, **kwargs)

    @classmethod
    def convert(self):
        if type == "tuple":
            tuple_vals = []
            for component in self.components:
                tuple_vals.append(component.convert())
            return EvmTuple(values=tuple_vals)


class SubscriptionResponse(TypedDict):
    id: int
    result: HexAddress
    jsorpc: VersionNumber


class EvmDataDict(TypedDict):
    address: HexAddress
    topics: list[HexStr]
    data: HexStr
    blockNumber: HexInt
    transactionHash: HexStr
    transactionIndex: HexInt
    blockHash: HexStr
    logIndex: HexInteger
    removed: bool


class EvmResult(TypedDict):
    result: EvmDataDict
    subscription: HexStr


class PendingResult(TypedDict):
    result: HexStr
    subscription: HexStr


class JsonPendingWssResponse(TypedDict):
    jsonrpc: VersionNumber
    method: str
    params: PendingResult
    subscription: HexStr


class JsonResponseWssResponse(TypedDict):
    jsonrpc: VersionNumber
    method: str
    params: EvmResult
    subscription: HexStr


__all__ = [
    "ALL_PRIMITIVES",
    "BASIC_TYPES",
    "BYTES_TYPES",
    "BLOCK_STRINGS",
    "BlockExplorer",
    "BlockReference",
    "Bytes32Hex",
    "Domain",
    "EIP712Model",
    "Indexed",
    "Name",
    "Network",
    "NoArgs",
    "RPCResponseModel",
    "Rpcs",
    "RpcUrl",
    "SignedTransaction",
    "Struct",
    "hash_eip712_bytes",
    "primitives",
]



from .account import GetAccountArgs
from .block import (
    AlchemyBlockReceipt,
    AlchemyParams,
    BlockNumberArg,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
)
from .eth_call import CallWithBlockArgs, EthCallArgs, EthCallParams
from .fee_history import FeeHistoryArgs
from .logs import LogsArgs, LogsParams
from .storage import GetCodeArgs, GetStorageArgs
from .tracer import TraceArgs, TracerConfig
from .transaction import (
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    RawTransaction,
    TransactionRequest,
)

__all__ = [
    "AlchemyBlockReceipt",
    "AlchemyParams",
    "BlockNumberArg",
    "CallWithBlockArgs",
    "EthCallArgs",
    "EthCallParams",
    "FeeHistoryArgs",
    "GetAccountArgs",
    "GetBlockByHashArgs",
    "GetBlockByNumberArgs",
    "GetCodeArgs",
    "GetStorageArgs",
    "GetTransactionByBlockHash",
    "GetTransactionByBlockNumber",
    "LogsArgs",
    "LogsParams",
    "RawTransaction",
    "TraceArgs",
    "TracerConfig",
    "TransactionRequest",
]



from inspect import isclass
from types import GenericAlias
from typing import Annotated, get_args, get_origin

from eth_abi import encode
from eth_account import Account
from eth_hash.auto import keccak
from eth_typing import ChecksumAddress, HexAddress, HexStr
from pydantic import Field

from .basic import ALL_PRIMITIVES
from .primitives import bytes32
from .struct import Struct


class EIP712Model(Struct):
    @classmethod
    def struct_name(cls) -> str:
        """Override this if you want the struct name to be different from the class name"""
        return cls.__name__

    @staticmethod
    def transform(type_):
        mapping = {
            str: "string",
            int: "uint256",  # defaults to uint256
            HexAddress: "address",
            ChecksumAddress: "address",
            HexStr: "bytes",
        }
        if type_ in mapping:
            return mapping[type_]
        if str(type_) in ALL_PRIMITIVES:
            return str(type_)
        # handle list type
        if isinstance(type_, GenericAlias):
            (arg,) = get_args(type_)
            return f"{arg.__name__}[]"
        return type_.__name__

    @classmethod
    def get_nested_types(cls):
        for _, field in cls.model_fields.items():
            for item in cls._get_nested_types(field.annotation):
                yield item

    @classmethod
    def _get_nested_types(cls, type_) -> list:
        if get_origin(type_) == list:
            return cls._get_nested_types(get_args(type_)[0])
        elif get_origin(type_) == tuple:
            args = get_args(type_)
            return [
                item
                for subl in [cls._get_nested_types(arg) for arg in args]
                for item in subl
            ]
        elif get_origin(type_) == Annotated:
            return cls._get_nested_types(get_args(type_)[0])
        elif isclass(type_) and issubclass(type_, EIP712Model):
            return [type_.type_string()] + list(type_.get_nested_types())
        return []

    @classmethod
    def type_string(cls) -> bytes:
        type_string: str = ""
        nested_types = list(cls.get_nested_types())
        for name, field_data in cls.model_fields.items():
            name = field_data.serialization_alias or name
            type_ = field_data.annotation

            print("TYPE::::", type_)
            if isinstance(type_, GenericAlias):
                print("??", type_)
                if get_origin(type_) == list:
                    (arg,) = get_args(type_)
                    type_str = f"{cls.transform(arg)}[]"
                elif get_origin(type_) == tuple:
                    args = get_args(type_)
                    type_str = f"({','.join([cls.transform(arg) for arg in args])})"
            else:
                type_str = cls.transform(type_)
            type_string += f"{type_str} {name},"
        encoded_typestr: bytes = (
            f'{cls.struct_name()}({type_string.rstrip(",")})'.encode("utf-8")
        )
        return encoded_typestr + b"".join(set(nested_types))

    @classmethod
    def typehash(cls):
        return keccak(cls.type_string())

    def hash(self) -> bytes32:
        types = ["bytes32"]
        values = [self.__class__.typehash()]
        for name, field_data in self.model_fields.items():
            type_ = field_data.annotation
            value = getattr(self, name)

            # handle a list of structs or primitive
            if isinstance(type_, GenericAlias):
                types.append("bytes32")

                (arg,) = get_args(type_)
                arg_name = self.transform(arg)

                if isclass(arg) and issubclass(arg, EIP712Model):
                    hashed_arr = []
                    for row in value:
                        hashed_arr.append(row.hash())
                    encoding = keccak(
                        hash_packed_array(
                            "bytes32",
                            hashed_arr,
                        )
                    )
                    values.append(encoding)
                else:
                    values.append(keccak(hash_packed_array(arg_name, value)))
            elif isclass(type_) and issubclass(type_, EIP712Model):
                types.append("bytes32")
                values.append(value.hash())
            else:
                type_name = self.transform(type_)
                if type_name == "string":
                    types.append("bytes32")
                    values.append(keccak(value.encode("utf-8")))
                elif type_name == "bytes":
                    types.append("bytes32")
                    values.append(keccak(value))
                else:
                    types.append(type_name)
                    values.append(value)
        return bytes32(keccak(encode(types, values)))

    def model_dump(self, by_alias=True, **kwargs):
        return super().model_dump(by_alias=by_alias, **kwargs)

    @staticmethod
    def recover(
        message: bytes,
        vrs: tuple[int, HexStr, HexStr] | None = None,
        signature: bytes | None = None,
    ):
        if vrs:
            return Account._recover_hash(message, vrs=vrs)
        return Account._recover_hash(message, signature=signature)


class Domain(EIP712Model):
    name: str
    version: str
    chain_id: int = Field(serialization_alias="chainId")
    verifying_contract: HexAddress = Field(serialization_alias="verifyingContract")

    @classmethod
    def struct_name(cls) -> str:
        return "EIP712Domain"


def hash_eip712_bytes(domain: Domain, struct: EIP712Model) -> bytes32:
    joined = b"\x19\x01" + domain.hash() + struct.hash()
    return bytes32(keccak(joined))


def hash_packed_array(type_, values) -> bytes:
    match type_:
        case "string":
            return encode(
                ["bytes32"] * len(values),
                [keccak(value.encode("utf-8")) for value in values],
            )
        case "bytes":
            return encode(
                ["bytes32"] * len(values),
                [keccak(value) for value in values],
            )
        case _:
            return encode(
                [type_] * len(values),
                values,
            )



from typing import Any

from eth_rpc.models import AccessListResponse, Account, FeeHistory, PendingTransaction
from eth_rpc.transaction import AlchemyReceiptsResponse
from eth_rpc.types import (
    AlchemyBlockReceipt,
    BlockNumberArg,
    CallWithBlockArgs,
    EthCallArgs,
    FeeHistoryArgs,
    GetAccountArgs,
    GetBlockByHashArgs,
    GetBlockByNumberArgs,
    GetCodeArgs,
    GetStorageArgs,
    GetTransactionByBlockHash,
    GetTransactionByBlockNumber,
    HexInteger,
    LogsArgs,
    NoArgs,
    RawTransaction,
    TraceArgs,
    TransactionRequest,
)
from eth_typing import HexStr

from ..block import Block
from ..log import Log
from ..transaction import Transaction as TransactionModel
from ..transaction import TransactionReceipt as TransactionReceiptModel
from .base import BaseRPC
from .method import RPCMethod


class RPC(BaseRPC):
    chain_id: RPCMethod = RPCMethod[NoArgs, HexInteger](name="eth_chainId")
    max_priority_fee_per_gas: RPCMethod = RPCMethod[NoArgs, HexInteger](
        name="eth_maxPriorityFeePerGas"
    )
    fee_history: RPCMethod = RPCMethod[FeeHistoryArgs, FeeHistory](
        name="eth_feeHistory"
    )
    eth_call: RPCMethod = RPCMethod[EthCallArgs, HexStr](name="eth_call")
    get_block_by_hash: RPCMethod = RPCMethod[GetBlockByHashArgs, Block](
        name="eth_getBlockByHash"
    )
    get_block_by_number: RPCMethod = RPCMethod[GetBlockByNumberArgs, Block](
        name="eth_getBlockByNumber"
    )
    get_block_tx_count_by_number: RPCMethod = RPCMethod[BlockNumberArg, HexInteger](
        name="eth_getBlockTransactionCountByNumber",
    )
    get_tx_by_hash: RPCMethod = RPCMethod[TransactionRequest, TransactionModel | None](
        name="eth_getTransactionByHash"
    )
    get_pending_tx_by_hash: RPCMethod = RPCMethod[
        TransactionRequest, PendingTransaction | None
    ](name="eth_getTransactionByHash")
    estimate_gas: RPCMethod = RPCMethod[CallWithBlockArgs, HexInteger](
        name="eth_estimateGas"
    )
    block_number: RPCMethod = RPCMethod[NoArgs, HexInteger](name="eth_blockNumber")
    create_access_list: RPCMethod = RPCMethod[CallWithBlockArgs, AccessListResponse](
        name="eth_createAccessList"
    )
    get_storage_at: RPCMethod = RPCMethod[GetStorageArgs, HexStr](
        name="eth_getStorageAt"
    )
    get_code: RPCMethod = RPCMethod[GetCodeArgs, HexStr](name="eth_getCode")
    get_tx_receipt: RPCMethod = RPCMethod[
        TransactionRequest, TransactionReceiptModel | None
    ](name="eth_getTransactionReceipt")
    get_block_receipts: RPCMethod = RPCMethod[
        list[HexInteger | HexStr], list[TransactionReceiptModel]
    ](name="eth_getBlockReceipts")
    alchemy_get_block_receipts: RPCMethod = RPCMethod[
        AlchemyBlockReceipt, AlchemyReceiptsResponse
    ](name="alchemy_getTransactionReceipts")
    get_tx_by_block_hash: RPCMethod = RPCMethod[
        GetTransactionByBlockHash, TransactionModel
    ](name="eth_getTransactionByBlockHashAndIndex")
    get_tx_by_block_number: RPCMethod = RPCMethod[
        GetTransactionByBlockNumber, TransactionModel
    ](name="eth_getTransactionByBlockNumberAndIndex")
    get_balance: RPCMethod = RPCMethod[GetAccountArgs, HexInteger](
        name="eth_getBalance"
    )
    get_account: RPCMethod = RPCMethod[GetAccountArgs, Account](name="eth_getAccount")
    get_tx_count: RPCMethod = RPCMethod[GetAccountArgs, HexInteger](
        name="eth_getTransactionCount"
    )
    get_logs: RPCMethod = RPCMethod[LogsArgs, list[Log]](name="eth_getLogs")

    send_raw_tx: RPCMethod = RPCMethod[RawTransaction, HexStr](
        name="eth_sendRawTransaction"
    )

    # debug methods
    debug_tracecall: RPCMethod = RPCMethod[TraceArgs, Any](name="debug_traceCall")

    def model_post_init(self, __context):
        rpc_methods = []
        for method in dir(self):
            if method[0] == "_":
                continue
            try:
                attr = getattr(self, method)
                if isinstance(attr, RPCMethod):
                    rpc_methods.append((method, attr))
            except Exception:
                pass

        for name, method in rpc_methods:
            new_method = method.copy()
            setattr(self, name, new_method.set_rpc(self).set_network(self.network))

        return super().model_post_init(__context)



import pytest
from eth_rpc import Account
from eth_rpc.networks import Arbitrum, Ethereum


@pytest.mark.unit
def test_account_network() -> None:
    # Currently, block is a singleton
    assert Account[Arbitrum]._network == Arbitrum
    assert Account[Ethereum]._network == Ethereum

    acct: type[Account[Arbitrum]] = Account[Arbitrum]
    assert acct._network == Arbitrum



import itertools
import time
from json import JSONDecodeError
from typing import TYPE_CHECKING, Generic, Optional, TypeVar

import httpx
from eth_rpc._response import RPCResponse
from eth_rpc.exceptions import RPCDecodeError
from eth_rpc.types import HexInt, HexStr, Network, NoArgs
from pydantic import BaseModel, ConfigDict, PrivateAttr

if TYPE_CHECKING:
    from ..core import RPC

Params = TypeVar(
    "Params", bound=BaseModel | HexInt | HexStr | list[HexInt | HexStr] | NoArgs
)
Response = TypeVar("Response")


class RPCMethodBase(BaseModel, Generic[Params, Response]):
    name: str
    client: Optional[httpx.AsyncClient] = None
    index: Optional[itertools.count] = None
    retries: int = 10
    _rpc: "RPC" = PrivateAttr(None)
    _network: type[Network] | None = PrivateAttr(None)

    def set_rpc(self, rpc: "RPC") -> "RPCMethodBase":
        self._rpc = rpc
        return self

    def set_network(self, network: Optional["Network"]) -> "RPCMethodBase":
        self._network = network
        return self

    def set_client(self, client: httpx.AsyncClient, index: itertools.count):
        self.client = client
        self.index = index

    def call_sync(self, *params: Params) -> Response:
        _, Output = self.__pydantic_generic_metadata__["args"]

        payload = {
            "method": self.name,
            "id": next(self._rpc.index),
            "jsonrpc": "2.0",
        }
        if not params:
            payload["params"] = []
        elif isinstance(params[0], HexInt):
            payload["params"] = hex(params[0])
        elif isinstance(params[0], str) or isinstance(params[0], list):
            payload["params"] = params[0]
        elif isinstance(params[0], BaseModel):
            payload["params"] = list(params[0].model_dump().values()) if params else []
        else:
            raise TypeError(f"Invalid Input Type: {type(params[0])}")
        while True:
            tries = 0
            try:
                response = self._send_sync(self._rpc, payload)
                break
            except (httpx.ReadTimeout, httpx.ConnectTimeout) as exc:
                tries += 1
                time.sleep(1)
                if tries == self._rpc.retries:
                    raise exc
        if "error" in response:
            raise ValueError(response["error"]["message"])
        return RPCResponse[Output](**response).result  # type: ignore

    async def call_async(self, *params: Params) -> Response:
        _, Output = self.__pydantic_generic_metadata__["args"]

        payload = {
            "method": self.name,
            "id": next(self._rpc.index),
            "jsonrpc": "2.0",
        }
        if not params or params == NoArgs:
            payload["params"] = []
        elif isinstance(params[0], HexInt):
            payload["params"] = hex(params[0])
        elif isinstance(params[0], str) or isinstance(params[0], list):
            # this is a HexStr or list of Hex Strings
            payload["params"] = params[0]
        elif isinstance(params[0], BaseModel):
            payload["params"] = list(params[0].model_dump().values()) if params else []

        response = await self._send_async(self._rpc, payload)
        if "error" in response:
            raise ValueError(response["error"]["message"])
        return RPCResponse[Output](**response, network=self._network).result  # type: ignore

    @staticmethod
    def _send_sync(rpc: "RPC", payload: dict) -> dict:
        result = httpx.post(rpc.http, json=payload, timeout=rpc.timeout)
        try:
            return result.json()
        except JSONDecodeError:
            raise RPCDecodeError(result.content)

    @staticmethod
    async def _send_async(rpc: "RPC", payload: dict) -> dict:
        result = await rpc.client.post(rpc.http, json=payload)
        try:
            return result.json()
        except JSONDecodeError:
            raise RPCDecodeError(result.content)

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )



from typing import ClassVar

from eth_rpc.types import BlockExplorer, Network, Rpcs, RpcUrl
from pydantic.networks import Url


class Base(Network):
    chain_id: ClassVar[int] = 8453
    name: ClassVar[str] = "Base"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://base.llamarpc.com"),
            wss=None,
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="Basescan",
        url="https://basescan.org",
        api_url="https://api.basescan.org/api",
    )
    alchemy_str: ClassVar[str | None] = "base-mainnet"
    apprx_block_time: ClassVar[float] = 2.0


class BaseSepolia(Network):
    chain_id: ClassVar[int] = 84532
    name: ClassVar[str] = "BaseSepolia"
    native_currency: ClassVar[str] = "ETH"
    rpc: ClassVar[Rpcs] = Rpcs(
        default=RpcUrl(
            http=Url("https://sepolia.base.org/"),
        )
    )
    block_explorer: ClassVar[BlockExplorer] = BlockExplorer(
        name="BaseScan-Sepolia",
        url="https://sepolia.basescan.org",
        api_url="https://api-sepolia.basescan.org/api",
    )
    alchemy_str: ClassVar[str] = "base-sepolia"
    apprx_block_time: ClassVar[float] = 2.0



from eth_rpc import PrivateKeyWallet
from eth_rpc.types import Domain, EIP712Model, hash_eip712_bytes, primitives
from eth_typing import HexStr
from pydantic import Field


class First(EIP712Model):
    x: int
    y: list[bool]


class Second(EIP712Model):
    first_: list[First] = Field(serialization_alias="first")
    second: tuple[bool, str]


def test_type_string():
    assert (
        Second.type_string()
        == b"Second(First[] first,(bool,string) second)First(uint256 x,bool[] y)"
    )


def test_domain_hash():
    domain = Domain(
        name="test",
        version="1",
        chain_id=1,
        verifying_contract="0x000000000000000000000000000000000000dead",
    )
    assert (
        domain.hash().hex()
        == "9304293a10b8d467c5e49bcf1cea604625f3771b366fee4aad461e904f7d9b8e"
    )

    assert (
        Second.type_string()
        == b"Second(First[] first,(bool,string) second)First(uint256 x,bool[] y)"
    )


def test_mail_example() -> None:
    # https://github.com/ethereum/EIPs/blob/master/assets/eip-712/Example.sol
    class Person(EIP712Model):
        name: str
        wallet: primitives.address

    class Mail(EIP712Model):
        from_: Person = Field(serialization_alias="from")
        to: Person
        contents: str

    mail = Mail(
        from_=Person(name="Cow", wallet="0xCD2a3d9F938E13CD947Ec05AbC7FE734Df8DD826"),
        to=Person(name="Bob", wallet="0xbBbBBBBbbBBBbbbBbbBbbbbBBbBbbbbBbBbbBBbB"),
        contents="Hello, Bob!",
    )

    domain = Domain(
        name="Ether Mail",
        version="1",
        chain_id=1,
        verifying_contract="0xCcCCccccCCCCcCCCCCCcCcCccCcCCCcCcccccccC",
    )
    assert (
        mail.type_string()
        == b"Mail(Person from,Person to,string contents)Person(string name,address wallet)"
    )
    assert domain.hash() == bytes.fromhex(
        "f2cee375fa42b42143804025fc449deafd50cc031ca257e0b194a650a912090f"
    )
    assert mail.hash() == bytes.fromhex(
        "c52c0ee5d84264471806290a3f2c4cecfc5490626bf912d01f240d7a274b371e"
    )

    hashed_mail = hash_eip712_bytes(domain, mail)

    v = 28
    r = HexStr("0x4355c47d63924e8a72e509b65029052eb6c299d53a04e167c5775fd466751c9d")
    s = HexStr("0x07299936d304c153f6443dfa05f40ff007d72911b6f72307f996231605b91562")
    assert EIP712Model.recover(hashed_mail, vrs=(v, r, s)) == mail.from_.wallet


def test_rvs():
    domain_hash = "9304293a10b8d467c5e49bcf1cea604625f3771b366fee4aad461e904f7d9b8e"
    wallet = PrivateKeyWallet(
        private_key="0xaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    )
    signature = wallet.sign_hash(domain_hash)
    assert (
        EIP712Model.recover(domain_hash, signature=signature.signature)
        == wallet.address
    )
    assert "0x" + signature.signature.hex() == wallet.rsv_to_signature(
        signature.r, signature.s, signature.v
    )
    wallet.signature_to_rsv(signature.signature.hex()) == (
        signature.r,
        signature.s,
        signature.v,
    )



from ._transport import (
    configure_rpc_from_env,
    get_current_network,
    set_alchemy_key,
    set_default_network,
    set_rpc_timeout,
    set_transport,
)
from .account import Account
from .block import Block
from .contract import Contract, ContractFunc, EthResponse, FuncSignature, ProtocolBase
from .event import Event
from .log import Log
from .models import EventData
from .subscriber import EventSubscriber
from .transaction import Transaction, TransactionReceipt
from .types import Network
from .wallet import PrivateKeyWallet

__all__ = [
    "Account",
    "Block",
    "Contract",
    "ContractFunc",
    "EthResponse",
    "Event",
    "EventData",
    "EventSubscriber",
    "FuncSignature",
    "Log",
    "Network",
    "PrivateKeyWallet",
    "ProtocolBase",
    "Transaction",
    "TransactionReceipt",
    "configure_rpc_from_env",
    "get_current_network",
    "set_alchemy_key",
    "set_default_network",
    "set_transport",
    "set_rpc_timeout",
]



from .address import address_to_topic, to_checksum
from .bloom import BloomFilter
from .datetime import convert_datetime_to_iso_8601, load_datetime_string
from .dual_async import handle_maybe_awaitable, run
from .model import RPCModel
from .streams import acombine, combine, ordered_iterator, sort_key
from .types import is_annotation, to_bytes32, to_hex_str, transform_primitive

__all__ = [
    "BloomFilter",
    "RPCModel",
    "run",
    "acombine",
    "address_to_topic",
    "to_bytes32",
    "combine",
    "convert_datetime_to_iso_8601",
    "handle_maybe_awaitable",
    "is_annotation",
    "load_datetime_string",
    "ordered_iterator",
    "sort_key",
    "to_checksum",
    "to_hex_str",
    "transform_primitive",
]



import pytest
from eth_rpc.types.basic import HexInt, HexInteger
from pydantic import BaseModel


@pytest.mark.unit
def test_hex_int() -> None:
    class MyNum(BaseModel):
        x: HexInt

    assert MyNum(x="0x7") == MyNum(x=7)
    my_num = MyNum(x=30)
    assert my_num.model_dump() == {"x": 30}


@pytest.mark.unit
def test_hex_integer() -> None:
    class MyNum(BaseModel):
        x: HexInteger

    assert MyNum(x="0x7") == MyNum(x=7)
    my_num = MyNum(x=30)
    assert my_num.model_dump() == {"x": "0x1e"}



import numbers
import operator

from eth_hash.auto import keccak as keccak_256


def get_chunks_for_bloom(value_hash):
    yield value_hash[:2]
    yield value_hash[2:4]
    yield value_hash[4:6]


def chunk_to_bloom_bits(chunk):
    high, low = bytearray(chunk)
    return 1 << ((low + (high << 8)) & 2047)


def get_bloom_bits(value):
    value_hash = keccak_256(value)
    for chunk in get_chunks_for_bloom(value_hash):
        bloom_bits = chunk_to_bloom_bits(chunk)
        yield bloom_bits


class BloomFilter(numbers.Number):
    value = None

    def __init__(self, value=0):
        self.value = value

    def __int__(self):
        return self.value

    def add(self, value):
        if not isinstance(value, bytes):
            raise TypeError("Value must be of type `bytes`")
        for bloom_bits in get_bloom_bits(value):
            self.value |= bloom_bits

    def extend(self, iterable):
        for value in iterable:
            self.add(value)

    @classmethod
    def from_iterable(cls, iterable):
        bloom = cls()
        bloom.extend(iterable)
        return bloom

    def __contains__(self, value):
        if not isinstance(value, bytes):
            raise TypeError("Value must be of type `bytes`")
        return all(self.value & bloom_bits for bloom_bits in get_bloom_bits(value))

    def __index__(self):
        return operator.index(self.value)

    def _combine(self, other):
        if not isinstance(other, (int, BloomFilter)):
            raise TypeError(
                "The `or` operator is only supported for other `BloomFilter` instances"
            )
        return BloomFilter(int(self) | int(other))

    def __or__(self, other):
        return self._combine(other)

    def __add__(self, other):
        return self._combine(other)

    def _icombine(self, other):
        if not isinstance(other, (int, BloomFilter)):
            raise TypeError(
                "The `or` operator is only supported for other `BloomFilter` instances"
            )
        self.value |= int(other)
        return self

    def __ior__(self, other):
        return self._icombine(other)

    def __iadd__(self, other):
        return self._icombine(other)

    def __hash__(self):
        return hash(self.value)



from collections.abc import Coroutine
from typing import Any, Literal, overload

import httpx
from httpx import Response

from .dual_async import DualAsync, MaybeAwaitable, run


class HTTPClient(DualAsync):
    async def _request(self, method, url, *args, sync: bool, **kwargs):
        if sync:
            return httpx.request(method, url, *args, **kwargs)
        async with httpx.AsyncClient() as client:
            return await client.request(method, url, *args, **kwargs)

    # GET
    @overload
    def get(self, url, *args, sync: Literal[True], **kwargs) -> Response: ...

    @overload
    def get(
        self, url, *args, sync: Literal[False] = ..., **kwargs
    ) -> Coroutine[Any, Any, Response]: ...

    def get(self, url, *args, sync: bool = False, **kwargs) -> MaybeAwaitable[Response]:
        return run(self._request, "GET", url, *args, sync=sync, **kwargs)

    # PUT
    @overload
    def put(self, url, *args, sync: Literal[True], **kwargs) -> Response: ...

    @overload
    def put(
        self, url, *args, sync: Literal[False] = ..., **kwargs
    ) -> Coroutine[Any, Any, Response]: ...

    def put(self, url, *args, sync: bool = False, **kwargs) -> MaybeAwaitable[Response]:
        return run(self._request, "PUT", url, *args, sync=sync, **kwargs)

    # POST
    @overload
    def post(self, url, *args, sync: Literal[True], **kwargs) -> Response: ...

    @overload
    def post(
        self, url, *args, sync: Literal[False] = ..., **kwargs
    ) -> Coroutine[Any, Any, Response]: ...

    def post(
        self, url, *args, sync: bool = False, **kwargs
    ) -> MaybeAwaitable[Response]:
        return run(self._request, "POST", url, *args, sync=sync, **kwargs)

    # DELETE
    @overload
    def delete(self, url, *args, sync: Literal[True], **kwargs) -> Response: ...

    @overload
    def delete(
        self, url, *args, sync: Literal[False] = ..., **kwargs
    ) -> Coroutine[Any, Any, Response]: ...

    def delete(
        self, url, *args, sync: bool = False, **kwargs
    ) -> MaybeAwaitable[Response]:
        return run(self._request, "DELETE", url, *args, sync=sync, **kwargs)



import os

import pytest
from eth_rpc import set_alchemy_key
from eth_rpc.ens import lookup_addr


@pytest.mark.asyncio(scope="session")
@pytest.mark.contract
async def test_ens_lookup():
    set_alchemy_key(os.environ["ALCHEMY_KEY"])
    response = await lookup_addr("vitalik.eth", block_number=20_000_000)
    assert response == "0xd8da6bf26964af9d7eed9e03e53415d37aa96045"

    response = await lookup_addr("000.eth", block_number=15_000_000)
    assert response == "0x9394a7cf1b7ad5de4825c3a3d8033453e23a4ce1"

    response = await lookup_addr("000.eth", block_number=20_000_000)
    assert response == "0xb34d849c0551c6a4c0d491224bcc6114b906afc4"



from .access_list import AccessList, AccessListResponse
from .account import Account
from .block import Block
from .fee_history import FeeHistory
from .log import EventData, Log
from .transaction import PendingTransaction, Transaction
from .transaction_receipt import TransactionReceipt

__all__ = [
    "AccessList",
    "AccessListResponse",
    "Account",
    "Block",
    "FeeHistory",
    "Log",
    "EventData",
    "Transaction",
    "PendingTransaction",
    "TransactionReceipt",
]



from typing import Any

from eth_typing import HexAddress, HexStr
from pydantic import BaseModel

from ..basic import BlockReference


class GetStorageArgs(BaseModel):
    storage_address: HexAddress
    slot_position: HexStr
    block_number: BlockReference


class GetCodeArgs(BaseModel):
    address: HexAddress
    block_number: BlockReference | None = None
    block_hash: HexStr | None = None

    def model_dump(self, exclude_none=True, **kwargs) -> dict[str, Any]:
        return super().model_dump(exclude_none=exclude_none, **kwargs)



from inspect import isclass
from typing import get_args, get_origin

from eth_abi import decode, encode
from pydantic import BaseModel


class Struct(BaseModel):
    """
    This is a type used to denote a solidity struct.  It is necessary
    to set this so the encoder knows to treat it as a single struct,
    and not a list of fields.

    For example, (uint8,bytes32) will encode differently from ((uint8,bytes32))

    One is a field with two arguments, the other is a field with a single argument,
        that is a struct with two fields
    """

    @classmethod
    def to_tuple_type(cls):
        lst = cls.to_type_list()
        return f"({','.join(lst)})"

    @classmethod
    def to_type_list(cls):
        from eth_rpc.utils.types import transform_primitive

        lst = []
        for _, field in cls.model_fields.items():
            type_ = field.annotation
            lst.append(transform_primitive(type_))
        return lst

    @classmethod
    def convert(cls, arg):
        if isinstance(arg, list):
            return [cls.convert(a) for a in arg]
        elif isinstance(arg, tuple):
            return tuple(cls.convert(a) for a in arg)
        elif isinstance(arg, Struct):
            return tuple(cls.convert(getattr(arg, a)) for a in arg.model_fields)
        return arg

    def to_bytes(self):
        types = self.to_tuple_type()
        values = []
        for _, value in self:
            values.append(self.convert(value))
        return encode([types], [tuple(values)])

    @classmethod
    def cast(cls, type_, args):
        if get_origin(type_) == list:
            (type_,) = get_args(type_)
            return [cls.cast(type_, arg) for arg in args]
        elif get_origin(type_) == tuple:
            tuple_types = get_args(type_)
            return tuple(cls.cast(type_, arg) for type_, arg in zip(tuple_types, args))
        elif isclass(type_) and issubclass(type_, Struct):
            fields = type_.model_fields
            model_dict = {}
            for (name, field), arg in zip(fields.items(), args):
                model_dict[name] = cls.cast(field.annotation, arg)
            return type_(**model_dict)
        return args

    @classmethod
    def from_bytes(cls, bytes_: bytes):
        types = cls.to_tuple_type()
        decoded = decode([types], bytes_)
        zipped = dict(zip(cls.model_fields.keys(), decoded[0]))
        fields = [field.annotation for field in cls.model_fields.values()]

        response = {}
        for field, (name, value) in zip(fields, zipped.items()):
            response[name] = cls.cast(field, value)
        return cls(**response)



from copy import deepcopy
from inspect import isclass
from typing import TYPE_CHECKING, ClassVar

from eth_rpc.types.args import EthCallParams, TraceArgs
from pydantic import BaseModel, PrivateAttr

from ._transport import _force_get_global_rpc
from .types import BlockReference, Network

if TYPE_CHECKING:
    from .rpc.core import RPC


class Request(BaseModel):
    _network: ClassVar[type[Network] | None] = PrivateAttr(default=None)

    def model_post_init(self, __context):
        network = self.__class__._network
        object.__setattr__(self, "_network", network)
        # overwrite the .rpc() classmethod
        object.__setattr__(self, "rpc", self._rpc)

    def __class_getitem__(cls, params):
        if isclass(params) and issubclass(params, Network):
            cls._network = params
            return cls
        else:
            try:
                return super().__class_getitem__(params)
            except AttributeError:
                pass
        return cls

    def __getitem__(self, params):
        if isclass(params) and issubclass(params, Network):
            if self._network != params:
                self = deepcopy(self)
                object.__setattr__(self, "_network", params)
        return self

    def _rpc(self) -> "RPC":
        """
        This uses the default network, unless a network has been provided
        """
        if self._network is None:
            return _force_get_global_rpc()
        response = _force_get_global_rpc(self._network)
        return response

    @classmethod
    def rpc(cls) -> "RPC":
        """
        This uses the default network, unless a network has been provided
        """
        from ._transport import _force_get_global_rpc

        if cls._network is None:
            return _force_get_global_rpc()
        response = _force_get_global_rpc(cls._network)
        return response

    def _get_debug_tracecall(
        self, address, data, block_number: BlockReference = "latest"
    ):
        return self.rpc().debug_tracecall(
            TraceArgs(
                params=EthCallParams(
                    to=address,
                    data=data,
                ),
                block_number=block_number,
            )
        )
